__all__ = [
    "FAString",
    "CatString",
]

from functools import partial, wraps
from inspect import signature
from typing import TYPE_CHECKING, List, NamedTuple, Optional, Union

if TYPE_CHECKING:
    from .rt_dataset import Dataset

import inspect
import re
import warnings
from functools import wraps

import numba as nb
import numpy as np
from numba.core.dispatcher import Dispatcher

from .config import get_global_settings
from .rt_enum import TypeRegister
from .rt_fastarray import FastArray
from .rt_numpy import empty, ones, unique, where, zeros
from .Utils.common import cached_property

# Partially-specialize the numba.njit decorator to simplify its use in the FAString class below.
_njit_serial = partial(nb.njit, parallel=False, cache=get_global_settings().enable_numba_cache, nogil=True)
_njit_par = partial(nb.njit, parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)


class FAStrDispatchPair(NamedTuple):
    """A pair of numba-based functions; one compiled for serial execution and the other for parallel execution."""

    serial: Dispatcher
    parallel: Dispatcher

    @staticmethod
    def create(py_func: callable) -> "FAStrDispatchPair":
        func_serial = _njit_serial(py_func)
        func_par = _njit_par(py_func)
        return FAStrDispatchPair(serial=func_serial, parallel=func_par)


def _warn_deprecated_naming(old_func, new_func):
    warnings.warn(
        f"`{old_func}` is now deprecated and has been renamed to `{new_func}`", DeprecationWarning, stacklevel=2
    )


# NOTE YOU MUST INSTALL tbb
# conda install tbb
# to confirm...
# >>> from numba import threading_layer()
# >>> threading_layer()
# >>> 'tbb'
#
# NOTE this class was first written reshaping the array as 2d
# ----
# if intype[1]=='S':
#    return self.view(np.uint8).reshape((len(self), self.itemsize))
# if intype[1]=='U':
#    return self.view(np.uint32).reshape((len(self), self.itemsize//4))
# ----
# then pass to numba as 2d using shape[0] outer loop, shape[1] inner loop with [row,col] accessors
# ---
# then flip back to string
# if self.itemsize == 4:
#    result=self.ravel().view('U'+str(self.shape[1]))
# else:
#    result=self.ravel().view('S'+str(self.shape[1]))
#
# Keeping it as 1d and remembering the itemsize is 10% faster on large arrays, and even faster on small arrays
# If recycling kicks in (only works on 1d arrays), it is 30% faster
# Also note: On Windows as of Aug 2019, prange is much slower without tbb installed
#
# subclass from FastArray


@nb.njit
def _str_equal(str1, str2):
    for char1, char2 in zip(str1, str2):
        if char1 != char2:
            return False
    return True


def _handle_apply_unique(func):
    sign = signature(func)

    if "apply_unique" not in sign.parameters:
        raise ValueError(f"apply_unique not found in the signature of {func}")

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        bound_args = sign.bind_partial(*args, **kwargs)
        bound_args.apply_defaults()

        if bound_args.kwargs["apply_unique"]:
            new_kwargs = bound_args.kwargs.copy()
            new_kwargs["apply_unique"] = False
            unique_values, index = unique(self.backtostring, return_inverse=True)
            result = func(unique_values.str, *bound_args.args, **new_kwargs)
            return result[index] if isinstance(result, FastArray) else result[index, :]
        else:
            return func(self, *args, **kwargs)

    return wrapper


def _maybe_encode(x):
    if not isinstance(x, bytes):
        x = x.encode()
    return x


class FAString(FastArray):
    """
    String accessor class for `FastArray`.
    """

    _APPLY_PARALLEL_THRESHOLD = 10_000

    def __new__(cls, arr):
        # if data comes in list like, convert to an array
        if not isinstance(arr, np.ndarray):
            if np.isscalar(arr):
                arr = np.asanyarray([arr])
            else:
                arr = np.asanyarray(arr)

        intype = arr.dtype.char
        if intype == "O":
            # try to convert to string (might have come from pandas)
            # default to unicode (note: FastArray attempts 'S' first)
            arr = arr.astype("U")
            intype = arr.dtype.char

        itemsize = np.int64(arr.itemsize)

        if intype == "S":
            # convert to two dim one byte array
            instance = arr.view(np.uint8)
        elif intype == "U":
            # convert to two dim four byte array
            instance = arr.view(np.uint32)
            itemsize = itemsize // 4
        else:
            raise TypeError(f"FAString can only be used on byte string and unicode not {arr.dtype!r}")

        instance = instance.view(cls)

        # remember the intype and itemsize
        instance._intype = intype
        instance._itemsize = itemsize
        return instance

    # -----------------------------------------------------
    @property
    def str(self):
        # we already are a str
        return self

    # -----------------------------------------------------
    @property
    def backtostring(self):
        """
        convert back to FastArray or np.ndarray 'S' or 'U' string
        'S12'  or 'U40'
        """
        return FastArray(self._np).view(self._intype + str(self._itemsize))

    @property
    def n_elements(self):
        """
        The number of elements in the original string array
        """
        return len(self) // self._itemsize

    # -----------------------------------------------------
    def possibly_convert_tostr(self, arr):
        """
        converts list like or an array to the same string type
        """

        # if data comes in list like, convert to an array
        if not isinstance(arr, np.ndarray):
            if not isinstance(arr, (list, tuple)):
                arr = np.asanyarray([arr])
            else:
                arr = np.asanyarray(arr)

        if arr.dtype.char != self._intype:
            arr = arr.astype(self._intype)

        return arr

    def _validate_input(self, str2):
        str2 = self.possibly_convert_tostr(str2)
        if len(str2) != 1:
            raise TypeError(f"A single string must be passed for str2 not {str2!r}")
        str2 = FAString(str2)
        return str2

    # -----------------------------------------------------
    def _apply_func(self, func, funcp, *args, dtype=None, input=None):
        # can optionally pass in dtype
        # check when to flip into parallel mode.  > 10,000 go to parallel routine
        # TODO: Can't just check len(self) for parallelism here; if operating on a string-mode Categorical
        #       the parallelism decision should be based on e.g. len(my_cat.category_array), not the length
        #       of the categorical itself.
        if len(self) >= self._APPLY_PARALLEL_THRESHOLD and funcp is not None:
            func = funcp
        if dtype is None:
            dest = empty(len(self), self.dtype)
            dest._itemsize = self._itemsize
            dest._intype = self._intype
        else:
            # user requested specific output dtype
            dest = empty(self.n_elements, dtype)

        if input is None:
            func(self._itemsize, dest, *args)
        else:
            if not isinstance(input, FAString):
                raise TypeError(f"The input= value was not a FAString.  It is {input!r}.")
            func(input, input._itemsize, dest, *args)

        if dtype is None:
            dest = dest.view(dest._intype + str(dest._itemsize))

        return dest

    # -----------------------------------------------------
    def apply(self, func, *args, dtype=None):
        """
        Write your own string apply function
        NOTE: byte strings are passed as uint8
        NOTE: unicode strings are passed as uint32

        default signature must match

        @nb.njit(cache=get_global_settings().enable_numba_cache, nogil=True)
        def nb_upper(src, itemsize, dest):

        src: is uint array
        itemsize: is how wide the string is per row
        dest: is return uint array

        Other Parameters
        ----------------
        *args: pass in zero or more arguments (the arguments are always at the end)
        dtype: specify a different dtype

        Example
        -------
        >>> import numba as nb
        ... @nb.njit(cache=get_global_settings().enable_numba_cache, nogil=True)
        ... def nb_upper(src, itemsize, dest):
        ...     for i in nb.prange(len(src) / itemsize):
        ...         rowpos = i * itemsize
        ...        for j in range(itemsize):
        ...             c=src[rowpos+j]
        ...             if c >= 97 and c <= 122:
        ...                 # convert to ASCII upper
        ...                 dest[rowpos+j] = c-32
        ...             else:
        ...                dest[rowpos+j] = c

        >>> FAString(['this  ','that ','test']).apply(nb_upper)

        """
        return self._apply_func(func, func, *args, dtype=dtype, input=self)

    # -----------------------------------------------------
    def _nb_upper_inplace(src, itemsize):
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            # loop over all chars in the string
            rowpos = i * itemsize
            # loop over all chars in the string
            for j in range(itemsize):
                c = src[rowpos + j]
                if c >= 97 and c <= 122:
                    # convert to ASCII upper
                    src[rowpos + j] = c - 32

    # -----------------------------------------------------
    def _nb_upper(src, itemsize, dest):
        # loop over all rows
        for i in nb.prange(np.int64(len(src) // itemsize)):
            rowpos = i * itemsize
            # loop over all chars in the string
            for j in range(itemsize):
                c = src[rowpos + j]
                if c >= 97 and c <= 122:
                    # convert to ASCII upper
                    dest[rowpos + j] = c - 32
                else:
                    dest[rowpos + j] = c

    # -----------------------------------------------------
    def _nb_lower(src, itemsize, dest):
        # loop over all rows
        for i in nb.prange(np.int64(len(src) // itemsize)):
            rowpos = i * itemsize
            # loop over all chars in the string
            for j in range(itemsize):
                c = src[rowpos + j]
                if c >= 65 and c <= 90:
                    # convert to ASCII lower
                    dest[rowpos + j] = c + 32
                else:
                    dest[rowpos + j] = c

    # -----------------------------------------------------
    def _nb_removetrailing(src, itemsize, dest, removechar):
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            # loop over all chars in the string backwards
            rowpos = i * itemsize

            startpos = -1
            for startpos in range(itemsize - 1, -1, -1):
                c = src[rowpos + startpos]
                is_trailing_char = c == 0 or c == removechar
                out_char = 0 if is_trailing_char else c
                dest[rowpos + startpos] = out_char

                if not is_trailing_char:
                    break

            for pos in range(startpos - 1, -1, -1):
                dest[rowpos + pos] = src[rowpos + pos]

    # -----------------------------------------------------
    def _nb_reverse_inplace(src, itemsize):
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            rowpos = i * itemsize
            # find length of string
            strlen = 0
            while strlen < itemsize:
                if src[rowpos + strlen] == 0:
                    break
                strlen += 1
            end = rowpos + strlen - 1
            start = rowpos
            while start < end:
                temp = src[end]
                src[end] = src[start]
                src[start] = temp
                start += 1
                end -= 1

    # -----------------------------------------------------
    def _nb_reverse(src, itemsize, dest):
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            rowpos = i * itemsize
            # find length of string
            strlen = 0
            while strlen < itemsize:
                if src[rowpos + strlen] == 0:
                    break
                strlen += 1
            srcpos = 0
            while strlen > 0:
                strlen -= 1
                dest[rowpos + strlen] = src[rowpos + srcpos]
                srcpos += 1
            while srcpos < itemsize:
                dest[rowpos + srcpos] = 0
                srcpos += 1

    # -----------------------------------------------------
    def _nb_strlen(src, itemsize, dest):
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            # loop over all chars in the string
            rowpos = i * itemsize
            strlen = 0
            # loop over all chars in the string
            for j in range(itemsize):
                if src[rowpos + j] == 0:
                    break
                strlen += 1
            # store length of string
            dest[i] = strlen

    # -----------------------------------------------------
    def _nb_index_any_of(src, itemsize, dest, str2):
        str2len = len(str2)
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            # loop over all chars in the string
            rowpos = i * itemsize
            found = 0
            # loop over all chars in the string
            for j in range(itemsize):
                c = src[rowpos + j]
                for k in range(str2len):
                    if c == str2[k]:
                        # store location of match
                        dest[i] = j
                        found = 1
                        break
                if found == 1:
                    break
            if found == 0:
                dest[i] = -1

    # -----------------------------------------------------
    def _nb_index(src, itemsize, dest, str2):
        str2len = len(str2)
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            # loop over all chars in the string
            rowpos = i * itemsize
            dest[i] = -1
            # loop over all substrings of sufficient length
            for j in range(itemsize - str2len + 1):
                if _str_equal(src[rowpos + j :], str2):
                    # store location of match
                    dest[i] = j
                    break

    # -----------------------------------------------------
    def _nb_contains(src, itemsize, dest, str2):
        str2len = len(str2)
        # loop over all rows
        for i in nb.prange(np.int64(len(src) // itemsize)):
            rowpos = i * itemsize
            dest[i] = False
            # loop over all substrings of sufficient length
            for j in range(itemsize - str2len + 1):
                if _str_equal(src[rowpos + j :], str2):
                    dest[i] = True
                    break

    def _nb_find(src, itemsize, dest, str2):
        """
        Searches src for occurrences of str2 and build a Boolean array
        with a row per string indicating indicating the starting points of all such occurrences.
        """
        str2len = len(str2)
        # loop over all rows
        for i in nb.prange(np.int64(len(src) // itemsize)):
            rowpos = i * itemsize
            # loop over all substrings of sufficient length
            str_pos = 0
            while str_pos <= itemsize - str2len:
                if _str_equal(src[rowpos + str_pos :], str2):
                    dest[i, str_pos] = True
                    str_pos += str2len
                else:
                    str_pos += 1

        return dest

    # -----------------------------------------------------

    def _nb_endswith(src, itemsize, dest, str2):
        str2len = len(str2)
        # loop over all rows
        for i in nb.prange(len(src) // itemsize):
            # loop over all chars in the string
            rowpos = i * itemsize
            dest[i] = False

            # loop over all chars in the string
            # check if enough space left
            if itemsize >= str2len:
                k = itemsize
                while (k > 0) and (src[rowpos + k - 1] == 0):
                    k -= 1

                # check if still enough space left
                if k >= str2len:

                    k2 = str2len
                    # check if only the end matches
                    while k2 > 0:
                        if src[rowpos + k - 1] != str2[k2 - 1]:
                            break
                        k -= 1
                        k2 -= 1
                    if k2 == 0:
                        # indicate we have a match
                        dest[i] = True

    # -----------------------------------------------------
    def _nb_startswith(src, itemsize, dest, str2):
        str2len = len(str2)
        # loop over all rows
        for i in nb.prange(np.int64(len(src) / itemsize)):
            # loop over all chars in the string
            rowpos = i * itemsize
            dest[i] = False
            # loop over all chars in the string
            # check if enough space left
            if itemsize >= str2len:
                k = 0
                # check if only the beginning matches
                while k < str2len:
                    if src[rowpos + k] != str2[k]:
                        break
                    k += 1
                if k == str2len:
                    # indicate we have a match
                    dest[i] = True

    # -----------------------------------------------------
    @property
    def upper(self):
        """
        upper case a string (bytes or unicode)
        makes a copy

        Examples
        --------
        >>> FAString(['this','that','test']).upper
        FastArray(['THIS','THAT','TEST'], dtype='<U4')

        """
        return self._apply_func(self.nb_upper, self.nb_upper_par)

    # -----------------------------------------------------
    @property
    def lower(self):
        """
        upper case a string (bytes or unicode)
        makes a copy

        Examples
        --------
        >>> FAString(['THIS','THAT','TEST']).lower
        FastArray(['this','that','test'], dtype='<U4')

        """
        return self._apply_func(self.nb_lower, self.nb_lower_par)

    # -----------------------------------------------------
    @property
    def upper_inplace(self):
        """
        upper case a string (bytes or unicode)
        does not make a copy

        Examples
        --------
        FAString(['this','that','test']).upper_inplace
        """
        # TODO: Enable parallel version + dispatching based on array length.
        self.nb_upper_inplace(self._itemsize)
        return self.backtostring

    # -----------------------------------------------------
    @property
    def reverse(self):
        """
        upper case a string (bytes or unicode)
        does not make a copy

        Examples
        --------
        FAString(['this','that','test']).reverse
        """
        return self._apply_func(self.nb_reverse, self.nb_reverse_par)

    # -----------------------------------------------------
    @property
    def reverse_inplace(self):
        """
        upper case a string (bytes or unicode)
        does not make a copy

        Examples
        --------
        FAString(['this','that','test']).reverse_inplace
        """
        # TODO: Enable parallel version + dispatching based on array length.
        self.nb_reverse_inplace(self._itemsize)
        return self.backtostring

    # -----------------------------------------------------
    def removetrailing(self, remove=32):
        """
        removes spaces at end of string (often to fixup matlab string)
        makes a copy

        Other Parameters
        ----------------
        remove=32.  defaults to removing ascii 32 (space character)

        Examples
        --------
        >>> FAString(['this  ','that ','test']).removetrailing()
        FastArray(['this','that','test'], dtype='<U6')
        """
        return self._apply_func(self.nb_removetrailing, self.nb_removetrailing_par, remove)

    # -----------------------------------------------------
    @cached_property  # only cached for Python 3.7 or higher
    def strlen(self):
        """
        return the string length of every string (bytes or unicode)

        Examples
        --------
        >>> FAString(['this  ','that ','test']).strlen
        FastArray([6, 5, 4])
        """
        return self._apply_func(self.nb_strlen, self.nb_strlen_par, dtype=np.int32)

    # -----------------------------------------------------
    def index_any_of(self, str2):
        """
        return the first index location any of the characters that are part of str2,
        or -1 if none of the characters match

        Parameters
        ----------
        str2 - a string with one or more characters to search for

        Examples
        --------
        >>> FAString(['this  ','that ','test']).index_any_of('ia')
        FastArray([2, 2, -1])
        """
        if not isinstance(str2, FAString):
            if str2 == "":
                return zeros(self.n_elements, dtype=np.int32)
            str2 = self._validate_input(str2)

        return self._apply_func(self.nb_index_any_of, self.nb_index_any_of_par, str2, dtype=np.int32)

    def strpbrk(self, str2):
        _warn_deprecated_naming("strpbrk", "index_any_of")
        return self.index_any_of(str2)

    # -----------------------------------------------------
    def index(self, str2):
        """
        return the first index location of the entire substring specified in str2,
        or -1 if the substring does not exist

        Parameters
        ----------
        str2 - a string with one or more characters to search for

        Examples
        --------
        >>> FAString(['this  ','that ','test']).index('at')
        FastArray([-1, 2, -1])
        """
        if not isinstance(str2, FAString):
            if str2 == "":
                return zeros(self.n_elements, dtype=np.int32)

            str2 = self._validate_input(str2)

        return self._apply_func(self.nb_index, self.nb_index_par, str2, dtype=np.int32)

    def strstr(self, str2):
        _warn_deprecated_naming("strstr", "index")
        return self.index(str2)

    # -----------------------------------------------------
    def contains(self, str2):
        """
        Return a boolean array that's True for each string element that contains the
        given substring, otherwise False.

        The entire substring must match.

        Parameters
        ----------
        str2 : str
            A string with one or more characters to search for. To search using regular
            expressions, use :meth:`FAString.regex_match`.

        Returns
        -------
        `FastArray`
            A boolean array where the value is True if the string contains the
            entire substring specified in `str2`, otherwise False.

        See Also
        --------
        FAString.startswith
        FAString.endswith
        FAString.regex_match

        Examples
        --------
        >>> FAString(['this  ','that ','test']).contains('at')
        FastArray([False, True, False])

        This can be called on a `FastArray` using ``.str.contains()``.

        >>> a = rt.FastArray(['this  ','that ','test'])
        >>> a.str.contains('at')
        FastArray([False,  True, False])
        """
        if not isinstance(str2, FAString):
            if str2 == "":
                return ones(self.n_elements, dtype=bool)

            str2 = self._validate_input(str2)

        return self._apply_func(self.nb_contains, self.nb_contains_par, str2, dtype=bool)

    def _find(self, str2):
        """
        Searches src for occurences of str2 and build a Boolean mask the same size
        as src indicating the starting point of all such occurences.

        Parameters
        ----------
        str2 - a string with one or more characters to search for

        Examples
        --------
        >>> FAString(['this','that','test']).find('t')
        FastArray([
            [True, False, False, False],
            [True, False, False, True],
            [True, False, False, True]
        ])
        """
        if not isinstance(str2, FAString):
            if str2 == "":
                return ones(len(self), dtype=bool)

            str2 = self._validate_input(str2)

        return self.nb_find(self._itemsize, str2=str2, dest=zeros((self.n_elements, self._itemsize), dtype=bool))

    def strstrb(self, str2):
        _warn_deprecated_naming("strstrb", "contains")
        return self.contains(str2)

    def _nb_replace(src, itemsize, dest, dest_itemsize, old, new, locations):
        old_len = len(old)
        new_len = len(new)

        new_is_empty = len(new) == 1 and new[0] == 0

        for row in nb.prange(np.int64(len(src) // itemsize)):
            rowpos = row * itemsize
            src_pos = 0
            dest_pos = row * dest_itemsize

            while src_pos < itemsize:
                if locations[row, src_pos]:
                    if new_is_empty:
                        src_pos += old_len
                    else:
                        dest[dest_pos : dest_pos + new_len] = new
                        src_pos += old_len
                        dest_pos += new_len
                else:
                    dest[dest_pos] = src[rowpos + src_pos]
                    src_pos += 1
                    dest_pos += 1
        return dest

    def replace(self, old: str, new: str) -> FastArray:
        """
        Replace all occurrences of `old` with `new`
        """
        if not isinstance(old, FAString):
            if old == "":
                raise ValueError("cannot replace the empty string")
            old = self._validate_input(old)

        new = self._validate_input(new)

        locations = self._find(old)
        if not locations.any():
            return self.backtostring

        char_diff = len(new) - len(old)
        if char_diff > 0:
            max_n_replacements = locations.sum(axis=1).max()
            dest_itemsize = self._itemsize + char_diff * max_n_replacements
        elif char_diff < 0:
            min_n_replacements = locations.sum(axis=1).min()
            dest_itemsize = self._itemsize - char_diff * min_n_replacements
        else:
            dest_itemsize = self._itemsize

        dest = zeros(self.n_elements * dest_itemsize, dtype=self.dtype)
        replace = self.nb_replace_par if len(self) >= self._APPLY_PARALLEL_THRESHOLD else self.nb_replace
        replaced = replace(
            itemsize=self._itemsize,
            dest=dest,
            dest_itemsize=dest_itemsize,
            old=old,
            new=new,
            locations=locations,
        )
        return replaced.view(self._intype + str(dest_itemsize))

    # -----------------------------------------------------
    def startswith(self, str2):
        """
        Return a boolean array that's True where the given substring matches the start
        of each string element, otherwise False.

        The entire substring must match.

        Parameters
        ----------
        str2 : str
            A string with one or more characters to search for. To search using regular
            expressions, use :meth:`FAString.regex_match`.

        Returns
        -------
        `FastArray`
            A boolean array where the value is True if the string starts with the
            entire substring specified in `str2`, otherwise False.

        See Also
        --------
        FAString.endswith
        FAString.contains
        FAString.regex_match

        Examples
        --------
        >>> FAString(['this  ','that ','test']).startswith('thi')
        FastArray([True, False, False])

        This can be called on a `FastArray` using ``.str.startswith()``.

        >>> a = rt.FastArray(['this  ','that ','test'])
        >>> a.str.startswith('thi')
        FastArray([True, False, False])
        """
        if not isinstance(str2, FAString):
            if str2 == "":
                return ones(self.n_elements, dtype=bool)

            str2 = self.possibly_convert_tostr(str2)
            if len(str2) != 1:
                return TypeError(f"A single string must be passed for str2 not {str2!r}")
            str2 = FAString(str2)

        return self._apply_func(self.nb_startswith, self.nb_startswith_par, str2, dtype=bool)

    # -----------------------------------------------------
    def endswith(self, str2):
        """
        Return a boolean array that's True where the given substring matches the end
        of each string element, otherwise False.

        The entire substring must match.

        Parameters
        ----------
        str2 : str
            A string with one or more characters to search for. To search using regular
            expressions, use :meth:`FAString.regex_match`.

        Returns
        -------
        `FastArray`
            A boolean array where the value is True if the string ends with the entire
            substring specified in `str2`, otherwise False.

        See Also
        --------
        FAString.startswith
        FAString.contains
        FAString.regex_match

        Examples
        --------
        >>> FAString(['abab','ababa','abababb']).endswith('ab')
        FastArray([True, False, False])

        This can be called on a `FastArray` using ``.str.endswith()``.

        >>> a = rt.FastArray(['abab','ababa','abababb'])
        >>> a.str.endswith('ab')
        FastArray([True, False, False])
        """
        if not isinstance(str2, FAString):
            if str2 == "":
                return ones(self.n_elements, dtype=bool)

            str2 = self.possibly_convert_tostr(str2)
            if len(str2) != 1:
                return TypeError(f"A single string must be passed for str2 not {str2!r}")
            str2 = FAString(str2)

        return self._apply_func(self.nb_endswith, self.nb_endswith_par, str2, dtype=bool)

    @_handle_apply_unique
    def regex_match(self, regex: Union["str", bytes], apply_unique: bool = True) -> FastArray:
        """
        Return a boolean array that's True where the given substring or regular
        expression pattern is contained in each string element, otherwise False.

        The entire substring or pattern must match.

        Applies :py:func:`re.search` on each element with `regex` as the pattern.

        Parameters
        ----------
        regex : str
            String or regular expression pattern to search for.
        apply_unique: bool, default True
            When True, the regex is applied to the unique values and then expanded
            using the reverse index (see :meth:`riptable.unique`). This is optimal
            for repetitive data and benign for unique or highly non-repetitive data.

        Returns
        -------
        `FastArray`
            A boolean array where the value is True if the string element contains the
            entire substring or regex pattern specified in `regex`, otherwise False.

        See Also
        --------
        FAString.regex_replace : Replace each instance of a specified substring or
            pattern.
        FAString.extract :
            Extract one or more pattern groups into a `Dataset` or `FastArray`.
        FAString.contains
        FAString.startswith
        FAString.endswith

        Examples
        --------
        Find any instance of 'ab' that appears at the end of a string:

        >>> FAString(['abab','ababa','abababb']).regex_match('ab$')
        FastArray([True, False, False])

        This can be called on a `FastArray` using ``.str.regex_match()``.

        >>> a = rt.FastArray(['abab','ababa','abababb'])
        >>> a.str.regex_match('ab$')
        FastArray([True, False, False])
        """
        if self._intype == "S":
            regex = _maybe_encode(regex)
        regex = re.compile(regex)
        vmatch = np.vectorize(lambda x: bool(regex.search(x)))
        bools = vmatch(self.backtostring)

        return bools

    @_handle_apply_unique
    def regex_replace(
        self, regex: Union["str", bytes], repl: Union["str", bytes], apply_unique: bool = True
    ) -> FastArray:
        """
        Replace each instance of a specified substring or pattern.

        The entire substring or pattern must match. If the substring or pattern isn't
        found, the original string is returned unchanged.

        The behavior is identical to that of :py:func:`re.sub`. In particular, the
        returned string is obtained by replacing the leftmost non-overlapping
        occurrences of the substring or pattern with the replacement string.

        Parameters
        ----------
        regex : str
            String or regular expression pattern to search for.
        repl : str
            The replacement string.
        apply_unique : bool, default True
            When True, the regex is applied to the unique values and then expanded
            using the reverse index (see :meth:`riptable.unique`). This is optimal
            for repetitive data and benign for unique or highly non-repetitive data.

        Returns
        -------
        `FastArray`
            An array with all occurrences of the substring or pattern replaced.

        See Also
        --------
        FAString.regex_match :
            Return a boolean array that indicates whether given substring or regular
            expression pattern is contained in each string element.
        FAString.extract :
            Extract one or more pattern groups into a `Dataset` or `FastArray`.
        FAString.contains
        FAString.startswith
        FAString.endswith

        Examples
        --------
        Replace instances of 'aa' with 'b'. All non-overlapping occurrences are
        replaced, starting from the left:

        >>> FAString(['aaa', 'aaaa', 'aaaaa']).regex_replace('aa', 'b')
        FastArray(['ba', 'bb', 'bba'], dtype='<U3>')

        Replace any instance of 'ab' that appears at the end of a string with 'b'.

        >>> FAString(['abab','ababa','abababb']).regex_replace('ab$', 'b')
        FastArray(['abb', 'ababa', 'abababb'], dtype='<U7')

        This can be called on a FastArray using ``.str.regex_replace()``. The
        returned `FastArray` elements are byte strings.

        >>> a = rt.FastArray(['abab','ababa','abababb'])
        >>> a.str.regex_replace('ab$', 'b')
        FastArray([b'abb', b'ababa', b'abababb'], dtype='|S7')
        """
        if self._intype == "S":
            regex, repl = map(_maybe_encode, (regex, repl))
        regex = re.compile(regex)
        vmatch = np.vectorize(lambda x: regex.sub(repl, x))
        return vmatch(self.backtostring)

    @_handle_apply_unique
    def extract(
        self, regex: str, expand: Optional[bool] = None, fillna: str = "", names=None, apply_unique: bool = True
    ) -> Union[FastArray, "Dataset"]:
        r"""
        Extract one or more pattern groups from each element of an array into a
        `FastArray` or `Dataset`.

        This is useful when you have pieces of data in a string that you want to split
        into separate elements.

        For one capture group, the default is to return a `FastArray`, but this can be
        overridden by setting `expand` to True or by providing a name of a `Dataset`
        column to populate. For more than one capture group, a `Dataset` is returned.

        Column names for the resulting `Dataset` can be specified within the regex
        using ``(?P<name>)`` in the capture group(s) or by passing the `names` argument,
        which may be more convenient.

        Parameters
        ----------
        regex : str
            The pattern(s) to search for. Define multiple capture groups using
            parentheses.
        expand : bool, default False
            Set to True to return a `Dataset` for a single capture group. If False, a
            `FastArray` is returned.
        fillna : str, default '' (empty string)
            For elements where there's no match, this is the fill value for the
            resulting `FastArray` or `Dataset` column.
        names : list of str, default None
            For more than one capture group, a `Dataset` is returned. Optionally, you
            can provide column names (keys) for the extracted data.
        apply_unique : bool
            When True, the regex is applied to the unique values and then expanded
            using the reverse index (see :meth:`riptable.unique`). This is optimal
            for repetitive data and benign for unique or highly non-repetitive data.

        Returns
        -------
        `FastArray` or `Dataset`
            For one capture group, a `FastArray` (or optionally a `Dataset`) is
            returned. For more than one capture group, a `Dataset` is returned.

        See Also
        --------
        FAString.regex_match :
            Return a boolean array that indicates whether given string or regular
            expression pattern is contained in each string element.
        FAString.regex_replace : Replace each instance of a specified string or pattern.

        Examples
        --------
        These examples use a `FastArray` containing OSI symbols.

        >>> osi = rt.FastArray(['SPX UO 12/15/23 C5700', 'SPXW UO 09/17/21 C3650'])

        Extract one substring:

        >>> osi.str.extract('\w+')
        FastArray([b'SPX', b'SPXW'], dtype='|S4')

        Provide a name for the resulting `Dataset` column:

        >>> osi.str.extract('(?P<root>\w+)')
        #   root
        -   ----
        0   SPX
        1   SPXW

        Define two capture groups and provide names for the resulting `Dataset` columns:

        >>> osi.str.extract('(\w+).* (\d{2}/\d{2}/\d{2})', names = ['root', 'expiration'])
        #   root   expiration
        -   ----   ----------
        0   SPX    12/15/23
        1   SPXW   09/17/21

        Extract one substring into a `Dataset` column using ``expand = True``. (Note
        that for the element with an unmatched pattern, an empty string is returned).

        >>> osi.str.extract('\w+W', expand = True)
        #   group_0
        -   -------
        0
        1   SPXW
        """
        kwargs = dict(expand=expand, fillna=fillna, names=names, apply_unique=apply_unique)

        if self._intype == "S":
            regex = _maybe_encode(regex)
        compiled = re.compile(regex)

        ngroups = compiled.groups
        if ngroups == 0:
            # convenience where we treat the entire pattern as a capture group
            return self.extract(f"({regex.decode()})", **kwargs)

        # expand defaults to False if we have one capture group and do not specify names
        if expand is None:
            expand = ngroups > 1 or names is not None or compiled.groupindex
        elif not expand and ngroups > 1:
            raise ValueError("expand cannot be False with multiple capture groups")

        if names is None:
            names = [f"group_{i}" for i in range(ngroups)]
            for name, index in compiled.groupindex.items():
                names[index - 1] = name
        elif len(names) != ngroups:
            raise ValueError(f"Number of names, {len(names)}, does not match number of groups, {ngroups}")

        strings = self.backtostring
        # we define a list containing an empty array for each group
        # use Python lists as we do not know how many chars will be in the resultant arrays.
        # Performance in comparable to unsing pre-allocated numpy arrays with ~600K unique elements
        out_arrs = [[fillna] * len(strings) for _ in range(ngroups)]

        for i, s in enumerate(strings):
            result = compiled.search(s)
            if result is not None:
                result = result.groups()
                for s, arr in zip(result, out_arrs):
                    arr[i] = s

        if expand:
            out = TypeRegister.Dataset(dict(zip(names, out_arrs)))
        else:
            out = FastArray(out_arrs[0])
        return out

    def _nb_substr(src, out, itemsize, start, stop, strlen):
        n_elements = len(out)
        max_chars = 0
        for elem in nb.prange(n_elements):
            elem_len = strlen[elem]
            i, j = start[elem], stop[elem]
            if i < 0:
                i += elem_len
            if j < 0:
                j += elem_len
            i = max(i, 0)
            j = min(j, elem_len)
            for out_pos, pos in enumerate(range(i, j)):
                char = src[itemsize * elem + pos]
                out[elem, out_pos] = char
                max_chars = max(max_chars, out_pos + 1)
        return out[:, :max_chars]

    def _substr(self, start: Union[int, np.ndarray], stop: Optional[Union[int, np.ndarray]] = None) -> FastArray:
        """
        Take a substring of each element using slice args.
        """
        if stop is None:
            # emulate behaviour of slice
            start, stop = 0, start

        def _bound_to_array(value, name):
            value = TypeRegister.FastArray(value)
            if len(value) == 1:
                value = np.full(self.n_elements, value[0])
            if value.shape != (self.n_elements,):
                raise ValueError(
                    f"{name} must be an integer or an array of length equal to the number of elements"
                    f" in self. Expected {(self.n_elements,)}, got {value.shape}"
                )
            return value

        start = _bound_to_array(start, "start")
        stop = _bound_to_array(stop, "stop")

        strlen = self.strlen

        out = zeros((self.n_elements, self._itemsize), self.dtype)

        if self.n_elements >= self._APPLY_PARALLEL_THRESHOLD:
            substr = self.nb_substr_par
        else:
            substr = self.nb_substr

        out = substr(out, self._itemsize, start, stop, strlen)

        n_chars = out.shape[1]
        if n_chars == 0:  # empty sub strings everywhere
            out = zeros(self.n_elements, self.dtype).view(f"{self._intype}1")
        else:
            out = out.ravel().view(f"<{self._intype}{n_chars}")
        return FastArray(out)

    @property
    def substr(self):
        return _SubStrAccessor(self)

    def substr_char_stop(self, stop: str, inclusive: bool = False) -> FastArray:
        """
        Take a substring of each element using characters as bounds.

        Parameters
        ----------
        stop:
            A string used to determine the start of the sub-string.
            Excluded from the result by default.
            We go to the end of the string where stop is not in found in the corresponding element
        inclusive: bool
            If True, include the stopping string in the result

        Examples
        --------
        >>> s = FastArray(['ABC', 'A_B', 'AB_C', 'AB_C_DD'])
        >>> s.str.substr_char_stop('_')
        FastArray([b'ABC', b'A', b'AB', b'AB'], dtype='|S2')
        >>> s.str.substr_char_stop('_', inclusive=True)
        FastArray([b'ABC', b'A_', b'AB_', b'AB_'], dtype='|S2')
        """
        int_stop = self.index(stop)
        int_stop[int_stop == -1] = self._itemsize  # return full string if stop not found
        if inclusive:
            int_stop += 1
        return self.substr(int_stop)

    def _nb_char(src, position, itemsize, strlen, out):
        broken_at = len(position)
        for i in nb.prange(len(position)):
            pos = position[i]
            if pos < 0:
                pos = strlen[i] + pos

            if pos >= itemsize or pos < 0:
                # Parallel reduction on this index.
                # Otherwise, returning here prevents the function from being parallelized.
                broken_at = np.minimum(broken_at, i)  # this triggers error below (in `char()`).

                # TODO: Set out[i] to some invalid value?
                out[i] = 0
            else:
                out[i] = src[itemsize * i + pos]

        return broken_at if broken_at < len(position) else -1

    def char(self, position: Union[int, List[int], np.ndarray]) -> FastArray:
        """
        Take a single character from each element.

        Parameters
        ----------
        position: int or list of int or np.ndarray
            The position of the character to be extracted. Negative values respect the
            length of the individual strings.
            If an array, the length must be equal to the number of strings.
            An error is raised if any positions are out of bounds (>= self._itemsize).
        """
        position = np.asanyarray(position)

        # Handle scalars
        if np.ndim(position) == 0:
            for size in [8, 16, 32, 64]:
                dtype = getattr(np, f"uint{size}")
                if self._itemsize <= np.iinfo(dtype).max:
                    break
            position = ones(self.n_elements, dtype) * position

        if len(position) != self.n_elements:
            raise ValueError("position must be a scalar or a vector of the same length as self")

        out = zeros(self.n_elements, self.dtype)
        broken_at = self._nb_char(position, self._itemsize, self.strlen, out)
        if broken_at >= 0:
            raise ValueError(f"Position {position[broken_at]} out of bounds " f"for string of length {self._itemsize}")
        out = out.view(f"{self._intype}1")
        return FastArray(out)

    # Use the specialized decorators to create both a serial and parallel version of each
    # numba function (so we only need one definition of each), then add it to FAString.
    nb_upper_inplace = _njit_serial(_nb_upper_inplace)
    nb_upper_inplace_par = _njit_par(_nb_upper_inplace)

    nb_upper = _njit_serial(_nb_upper)
    nb_upper_par = _njit_par(_nb_upper)

    # TODO: nb_lower_inplace

    nb_lower = _njit_serial(_nb_lower)
    nb_lower_par = _njit_par(_nb_lower)

    nb_removetrailing = _njit_serial(_nb_removetrailing)
    nb_removetrailing_par = _njit_par(_nb_removetrailing)

    nb_reverse_inplace = _njit_serial(_nb_reverse_inplace)
    nb_reverse_inplace_par = _njit_par(_nb_reverse_inplace)

    nb_reverse = _njit_serial(_nb_reverse)
    nb_reverse_par = _njit_par(_nb_reverse)

    nb_strlen = _njit_serial(_nb_strlen)
    nb_strlen_par = _njit_par(_nb_strlen)

    nb_index_any_of = _njit_serial(_nb_index_any_of)
    nb_index_any_of_par = _njit_par(_nb_index_any_of)

    nb_index = _njit_serial(_nb_index)
    nb_index_par = _njit_par(_nb_index)

    nb_contains = _njit_serial(_nb_contains)
    nb_contains_par = _njit_par(_nb_contains)

    nb_find = _njit_serial(_nb_find)
    nb_replace = _njit_serial(_nb_replace)
    nb_replace_par = _njit_serial(_nb_replace)

    nb_endswith = _njit_serial(_nb_endswith)
    nb_endswith_par = _njit_par(_nb_endswith)

    nb_startswith = _njit_serial(_nb_startswith)
    nb_startswith_par = _njit_par(_nb_startswith)

    nb_substr = _njit_serial(_nb_substr)
    nb_substr_par = _njit_par(_nb_substr)

    nb_char = _njit_serial(_nb_char)
    nb_char_par = _njit_par(_nb_char)


def _populate_wrappers(cls):
    """
    Decorator for the CatString class which populates the string methods and sets the defaults for
    filtered value fills.
    """
    functions = dict(inspect.getmembers(FAString, inspect.isfunction))
    properties = dict(inspect.getmembers(FAString, lambda o: isinstance(o, (cached_property, property))))
    for name in [
        "upper",
        "lower",
        "reverse",
        "removetrailing",
        "strlen",
        "index_any_of",
        "index",
        "contains",
        "startswith",
        "endswith",
        "regex_match",
        "regex_replace",
        "_substr",
        "char",
        "replace",
        "_find",
        # Deprecated methods.
        "strstr",
        "strstrb",
        "strpbrk",
    ]:
        if name in functions:
            wrapper = cls._build_method(functions[name])
        elif name in properties:
            wrapper = cls._build_property(name)
        else:
            raise RuntimeError(f"{name} is not defined on FAString as a function or property")
        setattr(cls, name, wrapper)
    return cls


class _SubStrAccessor:
    """
    Class for providing slicing on string arrays via FAString.substr
    """

    def __init__(self, fastring):
        self.fastring = fastring

    def __getitem__(self, y):
        if isinstance(y, slice):
            start = 0 if y.start is None else y.start
            return self.fastring._substr(start, y.stop)
        else:
            return self.fastring.char(y)

    def __call__(self, start, stop=None):
        return self.fastring._substr(start, stop)


@_populate_wrappers
class CatString:
    """
    Provides access to FAString methods for Categoricals.
    All string methods are wrappers of the FAString equivalent with
    categorical re-expansion and option for how to fill filtered elements.
    """

    def __init__(self, cat):
        from .rt_categorical import CategoryMode

        self.cat = cat
        if cat.category_mode == CategoryMode.StringArray:
            string_list = cat.category_array
            self.fastring = FAString(string_list)
        else:
            raise ValueError(f"Could not use .str in {CategoryMode(cat.category_mode).name}.")

    @cached_property
    def _isfiltered(self):
        return self.cat.isfiltered()

    def _convert_fastring_output(self, out):
        if out.dtype.kind in "SU":
            out = TypeRegister.Categorical._from_maybe_non_unique_labels(
                self.cat._fa, out, base_index=self.cat.base_index
            )
            return out
        else:
            return where(self._isfiltered, out.inv, out[self.cat.ikey - 1])

    @classmethod
    def _build_method(cls, method):
        """
        General purpose factory for FAString function wrappers.
        """

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            out = method(self.fastring, *args, **kwargs)
            return self._convert_fastring_output(out)

        return wrapper

    @classmethod
    def _build_property(cls, name):
        """
        General purpose factory for FAString property wrappers.
        """

        def wrapper(self):
            return self._convert_fastring_output(getattr(self.fastring, name))

        return property(wrapper)

    def extract(self, regex: str, expand: Optional[bool] = None, fillna: str = "", names=None):
        out = self.fastring.extract(regex, expand=expand, fillna=fillna, names=names, apply_unique=False)
        if isinstance(out, TypeRegister.Dataset):
            return TypeRegister.Dataset({key: self._convert_fastring_output(col) for key, col in out.items()})
        else:
            return self._convert_fastring_output(out)

    extract.__doc__ = FAString.__doc__  # might be misleading since we drop the apply_unique argument

    @property
    def substr(self):
        return _SubStrAccessor(self)


# keep as last line
TypeRegister.FAString = FAString
TypeRegister.CatString = CatString



import numpy as np

from ..rt_enum import (
    INVALID_DICT,
    INVALID_LONG_NAME,
    INVALID_SHORT_NAME,
    DisplayArrayTypes,
    DisplayColumnColors,
    DisplayJustification,
    DisplayLength,
    DisplayNumberSeparator,
    DisplayTextDecoration,
    NumpyCharTypes,
    TypeRegister,
)


# ---------------------------------------------------------------------
# ---------------------------------------------------------------------
class ItemFormat:
    """
    A container for display options for different data types in FastArray and numpy arrays.
    Basic numpy types have defaults. (see below)
    New types (subclassed from FastArray) will be queried to get formatting options.
    New types have the option of overwriting display_query_properties to set their own defaults.
    """

    def __init__(
        self,
        length=DisplayLength.Short,
        justification=DisplayJustification.Right,
        invalid=None,
        can_have_spaces=False,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
        convert=None,
        convert_format=None,
        format_string=None,
        timezone_str=None,
        maxwidth=None,
    ):

        self.length = length
        self.justification = justification
        self.invalid = invalid
        self.can_have_spaces = can_have_spaces
        self.html = html
        self.color = color
        self.decoration = decoration
        self.convert = convert
        self.convert_format = convert_format
        self.format_string = format_string
        self.timezone_str = timezone_str
        self.maxwidth = maxwidth

    # copy needs to be made for correct integer invalid
    def copy(self):
        new_format = ItemFormat(
            length=self.length,
            justification=self.justification,
            invalid=self.invalid,
            can_have_spaces=self.can_have_spaces,
            html=self.html,
            color=self.color,
            decoration=self.decoration,
            convert=self.convert,
            convert_format=self.convert_format,
            format_string=self.format_string,
        )
        return new_format

    # dump formatting options for debugging
    def __repr__(self):
        return self.summary()

    def __str__(self):
        return self.summary()

    def summary(self):
        info_str = ["\n     ItemFormat:"]
        info_str.append("         Length: " + str(self.length))
        info_str.append("  Justification: " + str(self.justification))
        info_str.append("        Invalid: " + str(self.invalid))
        info_str.append("Can have spaces: " + str(self.can_have_spaces))
        info_str.append("          Color: " + str(self.color))
        info_str.append("     Decoration: " + str(self.decoration))
        return "\n".join(info_str)


# ---------------------------------------------------------------------
default_item_formats = {
    DisplayArrayTypes.Bool: ItemFormat(
        length=DisplayLength.Long,
        justification=DisplayJustification.Right,
        invalid=None,
        can_have_spaces=False,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
    ),
    DisplayArrayTypes.Integer: ItemFormat(
        length=DisplayLength.Short,
        justification=DisplayJustification.Right,
        invalid=None,
        can_have_spaces=False,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
    ),
    DisplayArrayTypes.Float: ItemFormat(
        length=DisplayLength.Short,
        justification=DisplayJustification.Right,
        invalid=None,
        can_have_spaces=False,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
    ),
    DisplayArrayTypes.Bytes: ItemFormat(
        length=DisplayLength.Short,
        justification=DisplayJustification.Left,
        invalid=None,
        can_have_spaces=True,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
    ),
    DisplayArrayTypes.String: ItemFormat(
        length=DisplayLength.Short,
        justification=DisplayJustification.Left,
        invalid=None,
        can_have_spaces=True,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
    ),
    DisplayArrayTypes.Record: ItemFormat(
        length=DisplayLength.Short,
        justification=DisplayJustification.Left,
        invalid=None,
        can_have_spaces=True,
        html=False,
        color=DisplayColumnColors.Default,
        decoration=None,
    ),
}


# ---------------------------------------------------------------------
# ---------------------------------------------------------------------
class DisplayConvert:
    """
    Will analyze an array of a default type and return the appropriate conversion function.
    Anything that subclasses from FastArray will always fall back on the dtype of its underlying array.
    """

    convert_func_dict = {}

    # cache the array's dtype char to save type and conversion function
    # this will also be saved for any other occurrences of the same array type in a different dataset
    ConvertTypeCache = {}
    ConvertFuncCache = {}

    Verbose = 0

    def __init__(self):
        pass

    @staticmethod
    def get_display_convert(arr):
        if hasattr(arr, "dtype"):
            # check to see if the conversion function has been cached
            if arr.dtype in DisplayConvert.ConvertTypeCache:
                if DisplayConvert.Verbose > 1:
                    print("pulling conversion function from cache")
                arr_type = DisplayConvert.ConvertTypeCache[arr.dtype.char]
                convert_func = DisplayConvert.ConvertFuncCache[arr.dtype.char]
            else:
                arr_type = DisplayConvert.get_display_array_type(arr.dtype)
                convert_func = DisplayConvert.convert_func_dict.get(arr_type, DisplayConvert.convertDefault)
                # store the array type and conversion function into a cache so the lookup only happens once
                if DisplayConvert.Verbose > 1:
                    print("adding new conversion function to cache")
                DisplayConvert.ConvertTypeCache[arr.dtype.char] = arr_type
                DisplayConvert.ConvertFuncCache[arr.dtype.char] = convert_func
        else:
            # TODO: expand to handle python list-like objects
            arr_type = -1
            convert_func = DisplayConvert.convertDefault
        return arr_type, convert_func

    # ---------------------------------------------------------------------
    @staticmethod
    def get_display_array_type(dtype):
        """
        For FastArray and numpy array of basic type (not for objects that are subclasses of FastArray)
        """
        dtype_char = dtype.char
        if dtype_char == "?":
            return DisplayArrayTypes.Bool

        elif dtype_char in NumpyCharTypes.AllInteger:
            return DisplayArrayTypes.Integer

        elif dtype_char in NumpyCharTypes.AllFloat:
            return DisplayArrayTypes.Float

        elif dtype_char == "S":
            return DisplayArrayTypes.Bytes

        elif dtype_char == "U":
            return DisplayArrayTypes.String

        elif dtype_char == "V":
            return DisplayArrayTypes.Record

        # will use overall defaults
        else:
            return -1

    # ------------DEFAULT FORMATTING FUNCTIONS-----------------------------
    # ---------------------------------------------------------------------
    @staticmethod
    def convertBool(b, itemformat: ItemFormat):
        if DisplayConvert.Verbose > 0:
            print("***convert bool")
        result = str(b)
        if itemformat.length == DisplayLength.Long:
            return result
        else:
            return result[0]  # T/F

    # ---------------------------------------------------------------------
    @staticmethod
    def convertInt(i, itemformat: ItemFormat):
        if DisplayConvert.Verbose > 0:
            print("***convert int")
        if itemformat.invalid is not None:
            if i == itemformat.invalid:
                if DisplayConvert.Verbose > 0:
                    print("***found invalid int")
                if itemformat.length == DisplayLength.Short:
                    return INVALID_SHORT_NAME
                else:
                    return INVALID_LONG_NAME
        if TypeRegister.DisplayOptions.NUMBER_SEPARATOR:
            separator_string = "{:" + TypeRegister.DisplayOptions.NUMBER_SEPARATOR_CHAR + "}"
            return separator_string.format(i)
        else:
            return str(i)

    # ---------------------------------------------------------------------
    @staticmethod
    def convertFloat(f, itemformat: ItemFormat):
        if DisplayConvert.Verbose > 0:
            print("***convert float")
        # potential for long/short format
        separator = ""
        if TypeRegister.DisplayOptions.NUMBER_SEPARATOR:
            separator = TypeRegister.DisplayOptions.NUMBER_SEPARATOR_CHAR

        reg_precision = TypeRegister.DisplayOptions.PRECISION
        use_sci = False
        if f == f:
            f_test = abs(f)
            # lower bound for switching to scientific notation
            e_min = TypeRegister.DisplayOptions.e_min()
            # upper bound
            e_max = TypeRegister.DisplayOptions.e_max()
            if f_test != 0:
                if (f_test <= e_min) or (f_test >= e_max):
                    use_sci = True
                # switch to scientific notation if non-zero will appear as zero
                elif f_test < TypeRegister.DisplayOptions.p_threshold():
                    use_sci = True
        if use_sci:
            precision_str = "{:" + separator + "." + str(TypeRegister.DisplayOptions.E_PRECISION) + "e}"
        else:
            precision_str = "{:" + separator + "." + str(reg_precision) + "f}"

        f = precision_str.format(f)
        return f

    # ---------------------------------------------------------------------
    @staticmethod
    def convertBytes(i, itemformat: ItemFormat):
        if DisplayConvert.Verbose > 0:
            print("***convert bytes")
        # need to know long or short format
        # possibly enforce global maximum string
        if len(i) > 0:
            i = bytes.decode(i, errors="ignore")
            return trim_string(i, itemformat)
        else:
            return ""

    # ---------------------------------------------------------------------
    @staticmethod
    def convertString(i, itemformat: ItemFormat):
        if DisplayConvert.Verbose > 0:
            print("***convert string")
        return trim_string(i, itemformat)

    # ---------------------------------------------------------------------
    @staticmethod
    def convertDefault(i, itemformat: ItemFormat):
        if DisplayConvert.Verbose > 0:
            print("***convert default")
        i = str(i)
        return trim_string(i, itemformat)

    # ---------------------------------------------------------------------
    @staticmethod
    def convertMultiDims(i, itemformat: ItemFormat):
        """
        For displaying multi-dimensional arrays (currently only supports 2-dims).
        ItemFormat object contains a convert function for the dtype of the multidimensional array.
        """
        # for lots of dims, add lots of brackets
        final_list = []
        dims = i.ndim - 1
        # reduce to single dimension
        while i.ndim > 1:
            i = i[0]
        for item in i:
            final_list.append(itemformat.convert(item, itemformat))
        final_str = "[" * dims + str(final_list).replace("'", "")
        return final_str[: TypeRegister.DisplayOptions.MAX_STRING_WIDTH]

    # ---------------------------------------------------------------------
    @staticmethod
    def convertRecord(i, itemformat: ItemFormat):
        return str(i)


def trim_string(i, itemformat: ItemFormat):
    """
    If maxwidth was specified and is larger than the global default, it will be used instead.
    See also ColumnStyle()
    """
    if itemformat.maxwidth is not None and itemformat.maxwidth > TypeRegister.DisplayOptions.MAX_STRING_WIDTH:
        return i[: itemformat.maxwidth]
    if itemformat.length == DisplayLength.Short:
        i = i[: TypeRegister.DisplayOptions.MAX_STRING_WIDTH]
    return i


DisplayConvert.convert_func_dict = {
    DisplayArrayTypes.Bool: DisplayConvert.convertBool,
    DisplayArrayTypes.Integer: DisplayConvert.convertInt,
    DisplayArrayTypes.Float: DisplayConvert.convertFloat,
    DisplayArrayTypes.Bytes: DisplayConvert.convertBytes,
    DisplayArrayTypes.String: DisplayConvert.convertString,
    DisplayArrayTypes.Record: DisplayConvert.convertRecord,
}


def get_array_formatter(arr):
    """
    FastArray/subclasses have display_query_properties defined for custom string formatting.
    Numpy arrays have defaults in DisplayConvert.
    Returns ItemFormat object and display function for items in array based on type.
    """
    if hasattr(arr, "display_query_properties"):
        display_format, func = arr.display_query_properties()
    else:
        arr_type, func = DisplayConvert.get_display_convert(arr)
        display_format = default_item_formats.get(arr_type, ItemFormat())
    return display_format, func


def format_scalar(sc):
    """Convert a scalar to a string for display - the same way it would be if contained in a FastArray.
    Returns the converted scalar.
    """
    # consider writing a path just for scalars, wrapping in a FastArray
    arr = TypeRegister.FastArray(sc)
    display_format, func = get_array_formatter(arr)
    return func(arr[0], display_format)



import os

import pytest

from riptable import *
from riptable.rt_sds import SDSMakeDirsOn

# change to true since we write into /tests directory
SDSMakeDirsOn()


def arr_eq(a, b):
    return bool(np.all(a == b))


def arr_all(a):
    return bool(np.all(a))


class TestCategoricalFilterInvalid:
    def test_copy_new_filter_single(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        orig = FastArray([1, 1, 2, 3, 1])
        assert arr_eq(c._fa, orig)

        f = FastArray([False, True, True, True, True])
        d = Categorical(c, filter=f)
        new = FastArray([0, 1, 2, 3, 1])
        assert arr_eq(d._fa, new)

    def test_copy_new_filter_single2(self):
        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        orig = FastArray([1, 2, 3, 4, 5])
        assert arr_eq(c._fa, orig)

        f = FastArray([False, True, True, True, True])
        d = Categorical(c, filter=f)
        new = FastArray([0, 1, 2, 3, 4])
        assert arr_eq(d._fa, new)

    def test_copy_filter_errors(self):
        c = Categorical(["a", "a", "b", "c", "a"], base_index=0)
        f = FastArray([False, True, True, True, True])
        with pytest.warns(UserWarning):
            d = Categorical(c, filter=f)

    def test_copy_warnings(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        with pytest.warns(UserWarning):
            d = Categorical(c, ordered=False)

        with pytest.warns(UserWarning):
            d = Categorical(c, sort_gb=False)

        with pytest.warns(UserWarning):
            d = Categorical(c, lex=False)

        with pytest.warns(UserWarning):
            d = Categorical(c, base_index=0)

        with pytest.warns(UserWarning):
            d = Categorical(c, dtype=np.int64)

        with pytest.warns(UserWarning):
            d = Categorical(c, unicode=True)

        with pytest.warns(UserWarning):
            d = Categorical(c, invalid="inv")

    def test_filtered_set_name(self):
        c = Categorical([0, 1, 1, 2, 3], ["a", "b", "c"])
        assert c[0] == "Filtered"
        assert c.filtered_name == "Filtered"
        c.filtered_set_name("FILT")
        assert c[0] == "FILT"
        assert c.filtered_name == "FILT"

    # TODO move this into SDS save / load test module and use pytest fixtures
    def test_saveload_invalid(self):
        c = Categorical([0, 1, 1, 2, 3], ["a", "b", "c"], invalid="a")
        c.filtered_set_name("FILT")
        ds = Dataset({"catcol": c})
        ds.save(r"riptable/tests/temp/ds")
        ds2 = load_sds(r"riptable/tests/temp/ds")
        c2 = ds2.catcol
        assert c.invalid_category == c2.invalid_category
        assert c.filtered_name == c2.filtered_name
        os.remove(r"riptable/tests/temp/ds.sds")

    def test_isfiltered(self):
        c = Categorical(np.random.choice(4, 100), ["a", "b", "c"])
        flt = c.isfiltered()
        eq_z = c._fa == 0
        assert arr_eq(flt, eq_z)

    def test_isnan(self):
        c = Categorical(np.random.choice(4, 100), ["a", "b", "c"], invalid="a")
        inv = c.isnan()
        eq_bin = c._fa == 1
        assert arr_eq(inv, eq_bin)

    def test_combine_filter(self):
        c = Categorical(FA([2, 2, 1, 0, 2], dtype=np.int8), ["a", "b"])
        f = FA([True, True, False, True, False])
        combine_accum1_filter(c._fa, 2, filter=f)

    def test_single_key_filter(self):
        f = FA([True, True, False, True, False])
        pre_c = Categorical(["a", "a", "b", "c", "a"], filter=f)
        assert pre_c.unique_count == 2

        c = Categorical(["a", "a", "b", "c", "a"])
        assert c.unique_count == 3
        d = c.set_valid(filter=f)
        assert pre_c.unique_count == d.unique_count
        assert arr_eq(pre_c._fa, d._fa)

    def test_multikey_filter(self):
        f = FA([True, True, False, True, False])
        pre_c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)], filter=f)
        assert pre_c.unique_count == 3

        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        assert c.unique_count == 5
        d = c.set_valid(filter=f)
        assert pre_c.unique_count == d.unique_count
        assert arr_eq(pre_c._fa, d._fa)

    def test_categorical_empty(self):
        c = Categorical(zeros(10, dtype=np.int8), ["a", "b", "c"])
        assert c.unique_count == 3

        d = c.set_valid()
        assert d.unique_count == 0
        assert len(d.category_array) == 0

    def test_categorical_full(self):
        f = full(5, True)
        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        d = c.set_valid(filter=f)
        assert arr_eq(c._fa, d._fa)

    def test_filter_deep_copy(self):
        f = FA([True, True, False, True, False])
        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        assert c.unique_count == 5
        d = c.set_valid(filter=f)
        assert d.unique_count == 3
        assert c.unique_count == 5

    def test_filter_base_zero(self):
        f = FA([True, True, False, True, False])
        c = Categorical(["a", "a", "b", "c", "a"], filter=f)
        c_zero = Categorical(["a", "a", "b", "c", "a"], base_index=0)
        with pytest.warns(UserWarning):
            d = c_zero.set_valid(filter=f)
        assert arr_eq(c._fa, d._fa)
        assert c.unique_count == d.unique_count

    def test_pre_vs_post(self):
        # unique item removed
        arr = np.random.choice(["a", "b", "c"], 50)
        filter = arr != "b"
        c = Categorical(arr)
        c_pre = Categorical(arr, filter=filter)
        c_post = c.set_valid(filter=filter)
        c_copy = Categorical(c, filter=filter)

        assert arr_eq(c_pre.category_array, c_post.category_array)
        assert arr_eq(c_pre._fa, c_post._fa)

        assert arr_eq(c_pre.category_array, c_copy.category_array)
        assert arr_eq(c_pre._fa, c_copy._fa)

        # same uniques
        arr = np.random.choice(["a", "b", "c"], 50)
        filter = ones(50, dtype=bool)
        filter[:5] = False
        c = Categorical(arr)
        c_pre = Categorical(arr, filter=filter)
        c_post = c.set_valid(filter=filter)
        c_copy = Categorical(c, filter=filter)

        assert arr_eq(c_pre.category_array, c_post.category_array)
        assert arr_eq(c_pre._fa, c_post._fa)

        assert arr_eq(c_pre.category_array, c_copy.category_array)
        assert arr_eq(c_pre._fa, c_copy._fa)

    def test_expand_array_empty(self):
        arr = np.random.choice(["a", "b", "c"], 50)
        c = Categorical(arr)
        c2 = c.set_valid(filter=full(50, False))
        assert c2.unique_count == 0
        assert len(c2.category_array) == 0
        expanded = c2.expand_array
        assert arr_eq(expanded, c2.filtered_name)

    def test_expand_dict_empty(self):
        c = Categorical([arange(5), np.array(["a", "b", "c", "d", "e"])])
        c2 = c.set_valid(filter=full(5, False))
        assert c2.unique_count == 0

        d = list(c2.expand_dict.values())
        assert arr_all(d[0].isnan())
        assert arr_eq(d[1], c2.filtered_name)

    def test_gbc_multikey(self):
        names = FA(["brian", "brian", "adam", "charlie", "edward", "brian"])
        names.set_name("names")
        nums = FA([1, 1, 2, 3, 4, 1])
        nums.set_name("nums")

        f = FA([False, False, True, True, True, False])

        data = arange(6)

        c_multi = Categorical([names, nums])
        c_result = c_multi.sum(data)

        ds = Dataset({"names": names, "nums": nums})

        ds_result = ds.gbu(["names", "nums"]).sum(data)

        dsc = Dataset({"catcol": c_multi})
        dsc_result = dsc.gbu("catcol").sum(data)

        assert c_result.equals(ds_result), f"did not match for \n{c_result}\n{ds_result}"
        assert ds_result.equals(dsc_result), f"did not match for \n{ds_result}\n{dsc_result}"

    def test_enum_filter(self):
        codes = np.random.choice([10, 20, 30, 40], 50)
        d = {10: "aaa", 20: "bbb", 30: "ccc"}
        c = Categorical(codes, d)
        data = arange(50)

        # filter to go to groupby
        reg = c.sum(data)
        app = c.apply(sum, data)
        assert reg.equals(app)

        # keep enum as enum
        as_arr = c.set_valid(None)

        for i in range(len(as_arr)):
            assert c[i] == as_arr[i]

        # post filter enum
        codes = FA([10, 10, 20, 30, 10])
        d = {10: "aaa", 20: "bbb", 30: "ccc"}
        f = FA([True, True, False, True, True])
        c = Categorical(codes, d)

        c2 = c.as_singlekey().set_valid(f)
        assert c2.unique_count == 2
        c2.filtered_set_name("FLT")
        assert c2.unique_count == 2
        assert c2[2] == "FLT"

        c = Cat([10, 20, 30] * 3, {10: "A", 20: "B", 30: "C"})
        count = c.set_valid(c == "A").count()["Count"]
        assert np.all(count == [3, 6])

    def test_gbc_filter(self):
        pass

    def test_slice_empty_gb(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        c = c[:0]
        with pytest.raises(ValueError):
            _ = c.sum(arange(5))

        codes = FA([10, 10, 20, 30, 10])
        d = {10: "aaa", 20: "bbb", 30: "ccc"}
        c = Categorical(codes, d)
        c = c[:0]
        with pytest.raises(ValueError):
            _ = c.sum(arange(5))


class TestCategoricalSetValid:
    def test_copy_new_filter_single(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        orig = FastArray([1, 1, 2, 3, 1])
        assert arr_eq(c._fa, orig)

        f = FastArray([False, True, True, True, True])
        d = c.set_valid(f)
        new = FastArray([0, 1, 2, 3, 1])
        assert arr_eq(d._fa, new)

    def test_copy_new_filter_single2(self):
        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        orig = FastArray([1, 2, 3, 4, 5])
        assert arr_eq(c._fa, orig)

        f = FastArray([False, True, True, True, True])
        d = c.set_valid(f)
        new = FastArray([0, 1, 2, 3, 4])
        assert arr_eq(d._fa, new)

    def test_copy_filter_errors(self):
        c = Categorical(["a", "a", "b", "c", "a"], base_index=0)
        f = FastArray([False, True, True, True, True])
        with pytest.warns(UserWarning):
            d = c.set_valid(f)

    def test_single_key_filter(self):
        f = FA([True, True, False, True, False])
        pre_c = Categorical(["a", "a", "b", "c", "a"], filter=f)
        assert pre_c.unique_count == 2

        c = Categorical(["a", "a", "b", "c", "a"])
        assert c.unique_count == 3
        d = c.set_valid(filter=f)
        assert pre_c.unique_count == d.unique_count
        assert arr_eq(pre_c._fa, d._fa)

    def test_multikey_filter(self):
        f = FA([True, True, False, True, False])
        pre_c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)], filter=f)
        assert pre_c.unique_count == 3

        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        assert c.unique_count == 5
        d = c.set_valid(filter=f)
        assert pre_c.unique_count == d.unique_count
        assert arr_eq(pre_c._fa, d._fa)

    def test_categorical_empty(self):
        c = Categorical(zeros(10, dtype=np.int8), ["a", "b", "c"])
        assert c.unique_count == 3

        d = c.set_valid()
        assert d.unique_count == 0
        assert len(d.category_array) == 0

    def test_categorical_full(self):
        f = full(5, True)
        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        d = c.set_valid(filter=f)
        assert arr_eq(c._fa, d._fa)

    def test_filter_deep_copy(self):
        f = FA([True, True, False, True, False])
        c = Categorical([FA(["a", "a", "b", "c", "a"]), arange(5)])
        assert c.unique_count == 5
        d = c.set_valid(filter=f)
        assert d.unique_count == 3
        assert c.unique_count == 5

    def test_filter_base_zero(self):
        f = FA([True, True, False, True, False])
        c = Categorical(["a", "a", "b", "c", "a"], filter=f)
        c_zero = Categorical(["a", "a", "b", "c", "a"], base_index=0)
        with pytest.warns(UserWarning):
            d = c_zero.set_valid(filter=f)
        assert arr_eq(c._fa, d._fa)
        assert c.unique_count == d.unique_count

    def test_pre_vs_post(self):
        # unique item removed
        arr = np.random.choice(["a", "b", "c"], 50)
        filter = arr != "b"
        c = Categorical(arr)
        c_pre = Categorical(arr, filter=filter)
        c_post = c.set_valid(filter=filter)
        c_copy = Categorical(c, filter=filter)

        assert arr_eq(c_pre.category_array, c_post.category_array)
        assert arr_eq(c_pre._fa, c_post._fa)

        assert arr_eq(c_pre.category_array, c_copy.category_array)
        assert arr_eq(c_pre._fa, c_copy._fa)

        # same uniques
        arr = np.random.choice(["a", "b", "c"], 50)
        filter = ones(50, dtype=bool)
        filter[:5] = False
        c = Categorical(arr)
        c_pre = Categorical(arr, filter=filter)
        c_post = c.set_valid(filter=filter)
        c_copy = Categorical(c, filter=filter)

        assert arr_eq(c_pre.category_array, c_post.category_array)
        assert arr_eq(c_pre._fa, c_post._fa)

        assert arr_eq(c_pre.category_array, c_copy.category_array)
        assert arr_eq(c_pre._fa, c_copy._fa)

    def test_expand_array_empty(self):
        arr = np.random.choice(["a", "b", "c"], 50)
        c = Categorical(arr)
        c2 = c.set_valid(filter=full(50, False))
        assert c2.unique_count == 0
        assert len(c2.category_array) == 0
        expanded = c2.expand_array
        assert arr_eq(expanded, c2.filtered_name)

    def test_expand_dict_empty(self):
        c = Categorical([arange(5), np.array(["a", "b", "c", "d", "e"])])
        c2 = c.set_valid(filter=full(5, False))
        assert c2.unique_count == 0

        d = list(c2.expand_dict.values())
        assert arr_all(d[0].isnan())
        assert arr_eq(d[1], c2.filtered_name)

    def test_enum_filter(self):
        codes = np.random.choice([10, 20, 30, 40], 50)
        d = {10: "aaa", 20: "bbb", 30: "ccc"}
        c = Categorical(codes, d)
        data = arange(50)

        # filter to go to groupby
        reg = c.sum(data)
        app = c.apply(sum, data)
        assert reg.equals(app)

        # keep enum as enum
        as_arr = c.set_valid(None)

        for i in range(len(as_arr)):
            assert c[i] == as_arr[i]

        # post filter enum
        codes = FA([10, 10, 20, 30, 10])
        d = {10: "aaa", 20: "bbb", 30: "ccc"}
        f = FA([True, True, False, True, True])
        c = Categorical(codes, d)

        c2 = c.as_singlekey().set_valid(f)
        assert c2.unique_count == 2
        c2.filtered_set_name("FLT")
        assert c2.unique_count == 2
        assert c2[2] == "FLT"

        c = Cat([10, 20, 30] * 3, {10: "A", 20: "B", 30: "C"})
        count = c.set_valid(c == "A").count()["Count"]
        assert np.all(count == [3, 6])

    def test_slice_empty_gb(self):
        c = Categorical(["a", "a", "b", "c", "a"])
        c = c[:0]
        with pytest.raises(ValueError):
            _ = c.sum(arange(5))

        codes = FA([10, 10, 20, 30, 10])
        d = {10: "aaa", 20: "bbb", 30: "ccc"}
        c = Categorical(codes, d)
        c = c[:0]
        with pytest.raises(ValueError):
            _ = c.sum(arange(5))


# INVAID vs. FILTERED checklist:
# c = Cat(values)
# c = Cat(num values)
# c = Cat(values, cats)
# c = Cat(num values, num cats)
# ^^^^^^^
# AND all + filter


# test:
# isnan()
# isnotnan()
# isfiltered()
# isnotfiltered()

# For methods that fill with nan, fill with filtered bin 0, or fill with inv category bin?

# What if no invalid category specified?
# Is there a default for invalid category, or does it get set to None?
# Does numeric fall back on invalid for that dtype? Does string fall back on ''?
# How are filtered bins represented as string?

# Is the inv category being preserved during:
# -copy?
# -slice?
# -index?
# -ismember/mbget/hstack, etc.?
# -sds load (in meta data?)



__all__ = [
    "dataset_as_matrix",
    "numpy2d_to_dict",
    "dset_dict_to_list",
    "append_dataset_dict",
    "numpy_array_to_dict",
    "numpy_array_to_dataset",
]

from typing import TYPE_CHECKING

import numpy

import riptable as rt
from riptable.rt_datetime import UTC_1970_SPLITS, DateTimeNano

# TODO: Add more thorough type checking support, e.g, what to do with int64, uint64, datetime
# TODO: Should we convert to multiple columns for 64 bits, e.g, date, time-of-day for datetime?

if TYPE_CHECKING:
    from ..rt_dataset import Dataset


def _is_float_encodable(xtype: numpy.dtype):
    return numpy.issubdtype(xtype, numpy.integer) or numpy.issubdtype(xtype, numpy.floating)


def _normalize_column(x, field_key):
    original_type = x.dtype
    category_values = None
    is_categorical = False
    if _is_float_encodable(original_type):
        if isinstance(x, rt.rt_categorical.Categorical):
            category_values = x._categories
            is_categorical = True
        vals = x.astype(numpy.float64)
    else:
        if field_key is None:
            category_values, vals = numpy.unique(x, return_inverse=True)
            vals = vals.astype(numpy.float64)
        else:
            category_values = field_key
            isValid, vals = rt.ismember(x, category_values, 1)
            vals = vals.astype(numpy.float64)
            vals[~isValid] = numpy.nan
    return vals, original_type, is_categorical, category_values


def dataset_as_matrix(ds: "Dataset", save_metadata: bool = True, column_data={}):
    columns = list(ds.keys())
    nrows = ds.shape[0]
    ncols = ds.shape[1]  # TODO: may expand this for 64-bit columns
    out_array = numpy.empty((nrows, ncols), dtype=numpy.float64)
    column_info = dict()
    for col in range(ncols):
        field_key = column_data.get(columns[col])
        (
            out_array[:, col],
            original_type,
            is_categorical,
            category_values,
        ) = _normalize_column(ds[columns[col]], field_key)
        column_info[columns[col]] = {
            "dtype": original_type,
            "category_values": category_values,
            "is_categorical": is_categorical,
        }

    if save_metadata:
        return out_array, column_info
    else:
        return out_array


def numpy_array_to_dict(inarray: numpy.ndarray, columns=None):
    # TODO: Consolidate with the new `_possibly_convert_rec_array()` function in rt_utils.py
    #       that does some of this conversion in C++ and can take advantage of multi-threading.

    if len(inarray.shape) > 2:
        raise TypeError("Only 1 and 2-dimensional arrays are supported")
    out = dict()
    if inarray.dtype.fields is None:
        if len(inarray.shape) == 1:
            if columns is not None:
                if isinstance(columns, list):
                    if len(columns) > 1:
                        raise ValueError("Incompatible arrayShape and columns")
                    out[columns[0]] = inarray
                else:
                    if not isinstance(columns, str):
                        raise TypeError("Unexpected columns type")
                    out[columns] = inarray
            else:
                out["Var"] = inarray
            return out
        else:
            if columns is not None:
                if len(columns) != inarray.shape[1]:
                    raise ValueError("Incompatible arrayShape")
            else:
                columns = ["Var" + "_" + str(col_num) for col_num in range(inarray.shape[1])]
            for col_num in range(inarray.shape[1]):
                out[columns[col_num]] = inarray[:, col_num].copy()
            return out
    else:
        if len(inarray.shape) > 1:
            raise TypeError("Only 1-dimensional structured arrays are supported")
        if columns is not None:
            if len(inarray.dtype.fields) != len(columns):
                raise ValueError("Incompatible arrayShape")
        else:
            columns = [str(fld) for fld in inarray.dtype.fields]
        for col_num, fld in enumerate(inarray.dtype.fields):
            out[columns[col_num]] = inarray[fld].copy()
        return out


def numpy_array_to_dataset(inarray: numpy.ndarray, columns=None):
    out = rt.Dataset(numpy_array_to_dict(inarray, columns=columns))
    return out


# TODO: Replace all calls with numpy_array_to_dic? -CLH
def numpy2d_to_dict(arr, columns):
    """
    Converts arr 2D ndarray and column names to arr dict (is ordered) of ndarray's
    suitable for the Dataset constructor:

    :param arr: numpy NxM ndarray
    :param columns: list of M column names
    :return: dictionary suitable for rt.Dataset constructor

    Example:
    import numpy as np
    import riptable as rt
    from rt.Utils.conversion_utils import numpy2d_to_dict

    arr = np.array([[1, 2, 3], [4, 5, 6]])
    columns = ['c1', 'c2', 'c3']
    dset = rt.Dataset(numpy2d_to_dict(arr, columns))
    print(dset)
    """
    if len(columns) != arr.shape[1]:
        raise ValueError("Incompatible arrayShape={} and len(columns)={:d}".format(arr.shape, len(columns)))
    return dict(zip(columns, arr.T))


def dset_dict_to_list(ds_dict, key_field_name, allow_overwrite=False):
    """
    Converts a dict of Datasets to a list, appending the keyname as a new field key_field_name.
    NB: This modifies the Datasets!

    TODO: allow option of inplace or copy

    :param ds_dict: dictionary of Datasets.  Keys MUST be ascii strings (or bytes)!
    :param key_field_name: New column to add to each to which Dataset will be assigned
                           the constant value of the key.
    :param allow_overwrite: Unless set to True the key_field_name may not exist in any of the
                            input Datasets.
    :return: list of original Datasets _modified_.
    """
    if not allow_overwrite:
        for _d in ds_dict.values():
            if key_field_name in _d:
                raise ValueError(
                    "dset_dict_to_list(): key_field_name cannot be column name in any Dataset unless allow_overwrite=True."
                )
    max_length = 1
    for key in ds_dict:
        if type(key) is not bytes:
            try:
                _ = key.encode("ascii")
            except (AttributeError, UnicodeEncodeError):
                raise ValueError("dset_dict_to_list(): all keys must be ascii strings or bytes.")
        max_length = max(max_length, len(key))
    out = []
    typestr = f"|S{max_length}"
    for key, val in ds_dict.items():
        col_val = numpy.empty(shape=val.shape[0], dtype=typestr)
        col_val[:] = key
        # TODO: Replace with appropriate accessor
        val.__setattr__(key_field_name, col_val)
        out.append(val)
    return out


# -------------------------------------------------------
def append_dataset_dict(ds_dict, key_field_name):
    """
    Converts a dictionary of Datasets to a single Dataset appending the dictionary
    keys as key_field_name to distinguish them.
    NB: This modifies the original Datasets!

    TODO: add support for harmonizing fields, e.g.,
          fill missing values and deal with differing types

    :param ds_dict: dictionary of Datasets.  Keys MUST be ascii strings (or bytes)!
    :param key_field_name: New column to add to each to which Dataset will be assigned
                           the constant value of the key.
    :return: New dataset.
    """
    # check sort cache
    return rt.Dataset.concat_rows(dset_dict_to_list(ds_dict, key_field_name))


def possibly_convert_to_nanotime(vec):
    """
    Try to auto-detect nanotime and return DateTimeNano vector.

    :param vec: A numpy.ndarray, dtype numpy.int64.
    :return: DateTimeNano, True if successful, OrigVect, False if not.
    """
    # TODO: investigate reported problems with load
    if vec.dtype != numpy.int64:
        return vec, False
    # update this time window as necessary
    # print("**in possible nanotime")
    mintime = UTC_1970_SPLITS[44]
    maxtime = UTC_1970_SPLITS[49]
    # sample every 10
    arrlen = vec.size
    stepsize = arrlen // 10
    if arrlen < 10:
        sample = vec
    else:
        sample = vec[0:arrlen:stepsize]
    vmin = numpy.min(sample)
    vmax = numpy.max(sample)
    # if sampled data falls within 2015 - 2019 time window
    if vmin > mintime and vmax < maxtime:
        # print("**detected nano")
        return DateTimeNano(vec, from_tz="GMT", to_tz="NYC"), True
    # print("**failed nano")
    return vec, False
